from chellow.e.cfd import (
    _find_quarter_rs,
    _reconciled_quarters,
    hh,
    import_forecast_ilr_tra,
)
from chellow.e.computer import SupplySource
from chellow.models import (
    Comm,
    Contract,
    Cop,
    DtcMeterType,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcParticipant,
    Participant,
    Pc,
    Site,
    Source,
    VoltageLevel,
    insert_comms,
    insert_cops,
    insert_dtc_meter_types,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_reconciled_quarters(mocker):
    s = mocker.Mock()
    search_from = to_utc(ct_datetime(2023, 1, 1))
    log = mocker.Mock()
    result = [
        {"Settlement_Date": "2023-03-30", "Settlement_Run_Type": "DF"},
        {"Settlement_Date": "2023-03-30", "Settlement_Run_Type": "RF"},
        {"Settlement_Date": "2023-03-31", "Settlement_Run_Type": "DF"},
        {"Settlement_Date": "2023-03-31", "Settlement_Run_Type": "RF"},
    ]
    mocker.patch("chellow.e.cfd.api_records", return_value=result)
    actual = _reconciled_quarters(log, s, search_from)
    expected = {}
    assert actual == expected


def test_find_quarter_rs(sess):
    contract_name = "cfd_rates"
    vf = to_utc(ct_datetime(1996, 1, 1))
    participant = Participant.insert(sess, "CALB", "Calb")
    market_role = MarketRole.insert(sess, "Z", "Non-core")
    participant.insert_party(sess, market_role, "None core", vf, None, None)
    Contract.insert_non_core(
        sess,
        contract_name,
        "",
        {},
        to_utc(ct_datetime(2024, 1, 1)),
        None,
        {"rate_gbp_per_kwh": 0.10},
    )
    sess.commit()

    actual = _find_quarter_rs(sess, contract_name, to_utc(ct_datetime(2024, 4, 1)))
    assert actual is False


def test_import_forecast_ilr_tra(sess, mocker):
    vf = to_utc(ct_datetime(1996, 1, 1))
    participant = Participant.insert(sess, "CALB", "Calb")
    market_role = MarketRole.insert(sess, "Z", "Non-core")
    participant.insert_party(sess, market_role, "None core", vf, None, None)
    sess.commit()

    s = mocker.Mock()
    mock_request = mocker.Mock()
    req_j = {
        "success": True,
        "fields": [],
        "records": [],
    }
    mock_request.json = mocker.Mock(return_value=req_j)
    s.get = mocker.Mock(return_value=mock_request)
    log = mocker.Mock()
    set_progress = mocker.Mock()
    import_forecast_ilr_tra(sess, log, set_progress, s)


def test_import_forecast_ilr_tra_blank_lines(sess, mocker):
    vf = to_utc(ct_datetime(1996, 1, 1))
    participant = Participant.insert(sess, "CALB", "Calb")
    market_role = MarketRole.insert(sess, "Z", "Non-core")
    participant.insert_party(sess, market_role, "None core", vf, None, None)
    sess.commit()

    s = mocker.Mock()
    mock_request = mocker.Mock()
    req_j = {
        "success": True,
        "fields": [
            {"id": "_id", "type": "int"},
            {"id": "Quarterly_Obligation_Period", "type": "text"},
            {"id": "Period_Start", "type": "text"},
            {"id": "Period_End", "type": "text"},
            {"id": "Interim_Levy_Rate_GBP_Per_MWh", "type": "text"},
            {"id": "Total_Reserve_Amount_GBP", "type": "text"},
            {"id": "Eligible_Demand_MWh", "type": "text"},
            {"id": "BMRP_GBP_Per_MWh", "type": "text"},
            {"id": "IMRP_GBP_Per_MWh", "type": "text"},
            {"id": "Adjusted_ILR_GBP_Per_MWh", "type": "text"},
            {"id": "Additional_TRA_GBP", "type": "text"},
        ],
        "records": [[1, "", "", "", "", "", "", "", "", "", ""]],
    }
    mock_request.json = mocker.Mock(return_value=req_j)
    s.get = mocker.Mock(return_value=mock_request)
    log = mocker.Mock()
    set_progress = mocker.Mock()
    import_forecast_ilr_tra(sess, log, set_progress, s)


def test_hh(sess, mocker):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    participant = Participant.insert(sess, "CALB", "Calb")
    market_role = MarketRole.insert(sess, "Z", "Non-core")
    participant.insert_party(sess, market_role, "None core", vf, None, None)
    Contract.insert_non_core(
        sess, "cfd_in_period_tracking", "", {}, vf, None, {"rate_gbp_per_kwh": 0.10}
    )
    Contract.insert_non_core(sess, "cfd_forecast_ilr_tra", "", {}, vf, None, {})
    Contract.insert_non_core(
        sess,
        "cfd_advanced_forecast_ilr_tra",
        "",
        {},
        vf,
        None,
        {"sensitivity": {"Base Case": {"Interim Levy Rate_GBP_Per_MWh": "100"}}},
    )
    Contract.insert_non_core(
        sess,
        "cfd_reconciled_daily_levy_rates",
        "",
        {},
        vf,
        None,
        {"rate_gbp_per_kwh": 0.30},
    )
    Contract.insert_non_core(
        sess,
        "cfd_operational_costs_levy",
        "",
        {},
        vf,
        None,
        {"record": {"Effective_OCL_Rate_GBP_Per_MWh": 0.20}},
    )
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess, "bank_holidays", "", {}, vf, None, bank_holiday_rate_script
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    Contract.insert_dno(sess, dno.dno_code, participant, "", {}, vf, None, {})
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        vf,
        None,
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    sess.commit()
    caches = {}
    ss = SupplySource(
        sess,
        to_utc(ct_datetime(2020, 1, 1)),
        to_utc(ct_datetime(2020, 1, 1)),
        to_utc(ct_datetime(2021, 1, 1)),
        supply.eras[0],
        True,
        caches,
    )
    hh(ss)
    expected = {
        "anti-msp-kw": 0,
        "anti-msp-kwh": 0,
        "cfd-rates": {
            "interim": 0.1,
            "operational": 0.0002,
        },
        "ct-day": 1,
        "ct-day-of-week": 2,
        "ct-decimal-hour": 0.0,
        "ct-is-bank-holiday": False,
        "ct-is-month-end": False,
        "ct-month": 1,
        "ct-year": 2020,
        "exp-msp-kvar": 0,
        "exp-msp-kvarh": 0,
        "hist-export-grid-kvarh": 0,
        "hist-imp-msp-kvarh": 0,
        "hist-import-grid-kvarh": 0,
        "hist-kwh": 0,
        "hist-start": utc_datetime(2020, 1, 1, 0, 0),
        "imp-msp-kvar": 0,
        "imp-msp-kvarh": 0,
        "msp-kva": 0.0,
        "start-date": utc_datetime(2020, 1, 1, 0, 0),
        "status": "X",
        "utc-day": 1,
        "utc-day-of-week": 2,
        "utc-decimal-hour": 0.0,
        "utc-hour": 0,
        "utc-is-bank-holiday": False,
        "utc-is-month-end": False,
        "utc-minute": 0,
        "utc-month": 1,
        "utc-year": 2020,
        "msp-kwh": 0,
        "msp-kw": 0,
    }
    assert ss.hh_data[0] == expected


def test_hh_use_period(sess, mocker):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    participant = Participant.insert(sess, "CALB", "Calb")
    market_role = MarketRole.insert(sess, "Z", "Non-core")
    participant.insert_party(sess, market_role, "None core", vf, None, None)
    Contract.insert_non_core(
        sess,
        "cfd_in_period_tracking",
        "",
        {},
        vf,
        to_utc(ct_datetime(2020, 1, 1)),
        {"rate_gbp_per_kwh": 0.10},
    )
    Contract.insert_non_core(sess, "cfd_forecast_ilr_tra", "", {}, vf, None, {})
    Contract.insert_non_core(
        sess,
        "cfd_advanced_forecast_ilr_tra",
        "",
        {},
        vf,
        None,
        {"sensitivity": {"Base Case": {"Interim Levy Rate_GBP_Per_MWh": 100}}},
    )
    Contract.insert_non_core(
        sess,
        "cfd_reconciled_daily_levy_rates",
        "",
        {},
        vf,
        None,
        {"rate_gbp_per_kwh": 0.30},
    )
    Contract.insert_non_core(
        sess,
        "cfd_operational_costs_levy",
        "",
        {},
        vf,
        None,
        {"record": {"Effective_OCL_Rate_GBP_Per_MWh": 0.20}},
    )
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess, "bank_holidays", "", {}, vf, None, bank_holiday_rate_script
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    Contract.insert_dno(sess, dno.dno_code, participant, "", {}, vf, None, {})
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        vf,
        None,
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    sess.commit()
    caches = {}
    ss = SupplySource(
        sess,
        to_utc(ct_datetime(2020, 1, 1)),
        to_utc(ct_datetime(2020, 1, 1)),
        to_utc(ct_datetime(2021, 1, 1)),
        supply.eras[0],
        True,
        caches,
    )
    hh(ss)
    expected = {
        "anti-msp-kw": 0,
        "anti-msp-kwh": 0,
        "cfd-rates": {
            "interim": 0.1,
            "operational": 0.0002,
        },
        "ct-day": 1,
        "ct-day-of-week": 2,
        "ct-decimal-hour": 0.0,
        "ct-is-bank-holiday": False,
        "ct-is-month-end": False,
        "ct-month": 1,
        "ct-year": 2020,
        "exp-msp-kvar": 0,
        "exp-msp-kvarh": 0,
        "hist-export-grid-kvarh": 0,
        "hist-imp-msp-kvarh": 0,
        "hist-import-grid-kvarh": 0,
        "hist-kwh": 0,
        "hist-start": utc_datetime(2020, 1, 1, 0, 0),
        "imp-msp-kvar": 0,
        "imp-msp-kvarh": 0,
        "msp-kva": 0.0,
        "start-date": utc_datetime(2020, 1, 1, 0, 0),
        "status": "X",
        "utc-day": 1,
        "utc-day-of-week": 2,
        "utc-decimal-hour": 0.0,
        "utc-hour": 0,
        "utc-is-bank-holiday": False,
        "utc-is-month-end": False,
        "utc-minute": 0,
        "utc-month": 1,
        "utc-year": 2020,
        "msp-kwh": 0,
        "msp-kw": 0,
    }
    assert ss.hh_data[0] == expected


def test_hh_use_ilr(sess, mocker):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    participant = Participant.insert(sess, "CALB", "Calb")
    market_role = MarketRole.insert(sess, "Z", "Non-core")
    participant.insert_party(sess, market_role, "None core", vf, None, None)
    Contract.insert_non_core(
        sess, "cfd_in_period_tracking", "", {}, vf, None, {"rate_gbp_per_kwh": 0.10}
    )
    Contract.insert_non_core(
        sess,
        "cfd_forecast_ilr_tra",
        "",
        {},
        vf,
        to_utc(ct_datetime(2020, 1, 1)),
        {"record": {"Interim_Levy_Rate_GBP_Per_MWh": "100"}},
    )
    Contract.insert_non_core(
        sess,
        "cfd_advanced_forecast_ilr_tra",
        "",
        {},
        vf,
        None,
        {"sensitivity": {"Base Case": {"Interim Levy Rate_GBP_Per_MWh": 100}}},
    )
    Contract.insert_non_core(
        sess,
        "cfd_reconciled_daily_levy_rates",
        "",
        {},
        vf,
        None,
        {"rate_gbp_per_kwh": 0.30},
    )
    Contract.insert_non_core(
        sess,
        "cfd_operational_costs_levy",
        "",
        {},
        vf,
        None,
        {"record": {"Effective_OCL_Rate_GBP_Per_MWh": 0.20}},
    )
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess, "bank_holidays", "", {}, vf, None, bank_holiday_rate_script
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    Contract.insert_dno(sess, dno.dno_code, participant, "", {}, vf, None, {})
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        vf,
        None,
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    sess.commit()
    caches = {}
    ss = SupplySource(
        sess,
        to_utc(ct_datetime(2020, 1, 1)),
        to_utc(ct_datetime(2020, 1, 1)),
        to_utc(ct_datetime(2021, 1, 1)),
        supply.eras[0],
        True,
        caches,
    )
    hh(ss)
    expected = {
        "anti-msp-kw": 0,
        "anti-msp-kwh": 0,
        "cfd-rates": {
            "interim": 0.1,
            "operational": 0.0002,
        },
        "ct-day": 1,
        "ct-day-of-week": 2,
        "ct-decimal-hour": 0.0,
        "ct-is-bank-holiday": False,
        "ct-is-month-end": False,
        "ct-month": 1,
        "ct-year": 2020,
        "exp-msp-kvar": 0,
        "exp-msp-kvarh": 0,
        "hist-export-grid-kvarh": 0,
        "hist-imp-msp-kvarh": 0,
        "hist-import-grid-kvarh": 0,
        "hist-kwh": 0,
        "hist-start": utc_datetime(2020, 1, 1, 0, 0),
        "imp-msp-kvar": 0,
        "imp-msp-kvarh": 0,
        "msp-kva": 0.0,
        "start-date": utc_datetime(2020, 1, 1, 0, 0),
        "status": "X",
        "utc-day": 1,
        "utc-day-of-week": 2,
        "utc-decimal-hour": 0.0,
        "utc-hour": 0,
        "utc-is-bank-holiday": False,
        "utc-is-month-end": False,
        "utc-minute": 0,
        "utc-month": 1,
        "utc-year": 2020,
        "msp-kwh": 0,
        "msp-kw": 0,
    }
    assert ss.hh_data[0] == expected



from chellow.e.scenario import make_calcs, make_site_deltas
from chellow.models import (
    Comm,
    Contract,
    Cop,
    DtcMeterType,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcParticipant,
    Participant,
    Pc,
    Site,
    Source,
    VoltageLevel,
    insert_comms,
    insert_cops,
    insert_dtc_meter_types,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_make_site_deltas(mocker):
    era_1 = mocker.Mock()
    era_1.start_date = utc_datetime(2018, 1, 1)
    era_1.finish_date = None
    filter_returns = iter([[era_1], []])

    class Sess:
        def query(self, *args):
            return self

        def join(self, *args):
            return self

        def filter(self, *args):
            return next(filter_returns)

    sess = Sess()
    report_context = {}
    site = mocker.Mock()
    site.code = "1"
    scenario_hh = {site.code: {"used": "2019-03-01 00:00, 0"}}
    forecast_from = utc_datetime(2019, 4, 1)
    supply_id = None

    ss = mocker.patch("chellow.e.scenario.SiteSource", autospec=True)
    ss_instance = ss.return_value
    ss_instance.hh_data = [
        {
            "start-date": utc_datetime(2019, 3, 1),
            "used-kwh": 0,
            "export-grid-kwh": 0,
            "import-grid-kwh": 0,
            "msp-kwh": 0,
        }
    ]

    se = mocker.patch("chellow.e.scenario.SiteEra", autospec=True)
    se.site = mocker.Mock()

    sup_s = mocker.patch("chellow.e.scenario.SupplySource", autospec=True)
    sup_s_instance = sup_s.return_value
    sup_s_instance.hh_data = {}

    res = make_site_deltas(
        sess, report_context, site, scenario_hh, forecast_from, supply_id
    )

    assert len(res["supply_deltas"][False]["grid"]["site"]) == 0


def test_make_site_deltas_hh_multi_month(mocker):
    era_1 = mocker.Mock()
    era_1.start_date = utc_datetime(2018, 1, 1)
    era_1.finish_date = None
    filter_returns = iter([[era_1], []])

    class Sess:
        def query(self, *args):
            return self

        def join(self, *args):
            return self

        def filter(self, *args):
            return next(filter_returns)

    sess = Sess()
    report_context = {}
    site = mocker.Mock()
    site.code = "1"
    scenario_hh = {site.code: {"used": "2019-03-01 00:00, 0"}}
    forecast_from = utc_datetime(2019, 4, 1)
    supply_id = None

    ss = mocker.patch("chellow.e.scenario.SiteSource", autospec=True)
    ss_instance = ss.return_value
    ss_instance.hh_data = [
        {
            "start-date": utc_datetime(2019, 3, 1),
            "used-kwh": 0,
            "export-grid-kwh": 0,
            "import-grid-kwh": 0,
            "msp-kwh": 0,
        }
    ]

    se = mocker.patch("chellow.e.scenario.SiteEra", autospec=True)
    se.site = mocker.Mock()

    sup_s = mocker.patch("chellow.e.scenario.SupplySource", autospec=True)
    sup_s_instance = sup_s.return_value
    sup_s_instance.hh_data = {}

    res = make_site_deltas(
        sess, report_context, site, scenario_hh, forecast_from, supply_id
    )
    filter_returns = iter([[era_1], []])

    class Sess:
        def query(self, *args):
            return self

        def join(self, *args):
            return self

        def filter(self, *args):
            return next(filter_returns)

    sess = Sess()
    res = make_site_deltas(
        sess, report_context, site, scenario_hh, forecast_from, supply_id
    )

    assert len(res["supply_deltas"][False]["grid"]["site"]) == 0


def test_make_site_deltas_nhh(mocker):
    era_1 = mocker.Mock()
    era_1.start_date = utc_datetime(2018, 1, 1)
    era_1.finish_date = None
    """
    filter_args = iter(
        [
            [
                'False',
                'era.finish_date IS NULL OR era.finish_date >= :finish_date_1',
                'era.imp_mpan_core IS NOT NULL',
                'era.start_date <= :start_date_1', 'pc.code != :code_1',
                'true = :param_1'
            ],
            [
                'False',
                'era.finish_date IS NULL OR era.finish_date >= :finish_date_1',
                'era.imp_mpan_core IS NOT NULL',
                'era.start_date <= :start_date_1', 'source.code = :code_1',
                'true = :param_1'
            ]
        ]
    )
    """

    filter_returns = iter([[era_1], []])

    class Sess:
        def query(self, *args):
            return self

        def join(self, *args):
            return self

        def filter(self, *args):
            """
            actual_args = sorted(map(str, args))
            expected_args = next(filter_args)
            assert actual_args == expected_args
            """
            return next(filter_returns)

    sess = Sess()
    report_context = {}
    site = mocker.Mock()
    site.code = "1"
    scenario_hh = {site.code: {"used": "2019-03-01 00:00, 0"}}
    forecast_from = utc_datetime(2019, 4, 1)
    supply_id = None

    ss = mocker.patch("chellow.e.scenario.SiteSource", autospec=True)
    ss_instance = ss.return_value
    ss_instance.hh_data = [
        {
            "start-date": utc_datetime(2019, 3, 1),
            "used-kwh": 0,
            "export-grid-kwh": 0,
            "import-grid-kwh": 0,
            "msp-kwh": 0,
        }
    ]

    se = mocker.patch("chellow.e.scenario.SiteEra", autospec=True)
    se.site = mocker.Mock()

    sup_s = mocker.patch("chellow.e.scenario.SupplySource", autospec=True)
    sup_s_instance = sup_s.return_value
    hh_start_date = utc_datetime(2019, 3, 1)
    sup_s_instance.hh_data = [{"start-date": hh_start_date, "msp-kwh": 10}]

    res = make_site_deltas(
        sess, report_context, site, scenario_hh, forecast_from, supply_id
    )

    assert res["supply_deltas"][True]["grid"]["site"] == {hh_start_date: -10.0}


def test_make_calcs_new_generation(mocker, sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    start_date = utc_datetime(2009, 7, 31, 23, 00)
    finish_date = utc_datetime(2009, 8, 31, 22, 30)
    supply_ids = None
    report_context = {}
    forecast_from = utc_datetime(2020, 1, 1)

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        vf,
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_charge_script = """
import chellow.e.ccl
from chellow.utils import HH, reduce_bill_hhs, utc_datetime

def virtual_bill_titles():
    return [
        'ccl-kwh', 'ccl-rate', 'ccl-gbp', 'net-gbp', 'vat-gbp', 'gross-gbp',
        'sum-msp-kwh', 'sum-msp-gbp', 'problem']

def virtual_bill(ds):
    for hh in ds.hh_data:
        hh_start = hh['start-date']
        bill_hh = ds.supplier_bill_hhs[hh_start]
        bill_hh['sum-msp-kwh'] = hh['msp-kwh']
        bill_hh['sum-msp-gbp'] = hh['msp-kwh'] * 0.1
        bill_hh['net-gbp'] = sum(
            v for k, v in bill_hh.items() if k.endswith('gbp'))
        bill_hh['vat-gbp'] = 0
        bill_hh['gross-gbp'] = bill_hh['net-gbp'] + bill_hh['vat-gbp']

    ds.supplier_bill = reduce_bill_hhs(ds.supplier_bill_hhs)

def displaced_virtual_bill(ds):
    for hh in ds.hh_data:
        hh_start = hh['start-date']
        bill_hh = ds.supplier_bill_hhs[hh_start]
        bill_hh['sum-msp-kwh'] = hh['msp-kwh']
        bill_hh['sum-msp-gbp'] = hh['msp-kwh'] * 0.1
        bill_hh['net-gbp'] = sum(
            v for k, v in bill_hh.items() if k.endswith('gbp'))
        bill_hh['vat-gbp'] = 0
        bill_hh['gross-gbp'] = bill_hh['net-gbp'] + bill_hh['vat-gbp']

    ds.supplier_bill = reduce_bill_hhs(ds.supplier_bill_hhs)
"""
    imp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        imp_supplier_charge_script,
        {},
        vf,
        None,
        {},
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    Contract.insert_dno(
        sess,
        dno.dno_code,
        participant,
        "",
        {},
        vf,
        None,
        {},
    )
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", True, False, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc_imp = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        vf,
        None,
    )
    dno.insert_llfc(
        sess,
        "521",
        "Export (HV)",
        voltage_level,
        False,
        False,
        vf,
        None,
    )
    MtcLlfc.insert(sess, mtc_participant, llfc_imp, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )

    sess.commit()

    scenario_hh = {
        "CI017": {
            "generated": """
                2009-08-01 00:00, 40
                2009-08-15 00:00, 40"""
        }
    }

    era_maps = {
        utc_datetime(2000, 8, 1): {
            "llfcs": {"22": {"new_export": "521"}},
            "supplier_contracts": {"new_export": 4},
        }
    }

    site_deltas = make_site_deltas(
        sess, report_context, site, scenario_hh, forecast_from, supply_ids
    )
    calcs, _ = make_calcs(
        sess,
        site,
        start_date,
        finish_date,
        supply_ids,
        site_deltas,
        forecast_from,
        report_context,
        era_maps,
        None,
    )
    assert calcs[1][1] == "displaced"
    assert calcs[2][1] == "CI017_extra_gen_TRUE"
    assert calcs[3][2] == "CI017_extra_grid_export"



from decimal import Decimal

from openpyxl import Workbook

from chellow.e.bill_parsers.edf_export_xlsx import _make_raw_bills
from chellow.utils import ct_datetime, to_utc


def test_make_raw_bills():
    wb = Workbook()
    statement_sheet = wb.worksheets[0]
    wb.create_sheet()  # HH Data
    wb.create_sheet()  # Calculations
    gduos_sheet = wb.create_sheet()  # GDUoS

    statement_sheet["D22"] = "01-May-2024"
    statement_sheet["F22"] = "31-May-2024"
    statement_sheet["H24"] = ct_datetime(2024, 8, 4)
    statement_sheet["D25"] = 1000000000003
    statement_sheet["D33"] = Decimal("302")
    statement_sheet["H33"] = Decimal("23.09")

    gduos_sheet["A2"] = 1000000000003
    gduos_sheet["B2"] = "2024-04"
    gduos_sheet["D2"] = "Capacity"
    gduos_sheet["E2"] = Decimal("41")
    gduos_sheet["F2"] = Decimal("4.97")
    gduos_sheet["G2"] = Decimal("83.12")

    bills = _make_raw_bills(wb)
    expected_bills = [
        {
            "account": "10 0000 0000 003",
            "bill_type_code": "N",
            "breakdown": {"ssp-gbp": Decimal("23.09"), "ssp-kwh": Decimal("302")},
            "finish_date": to_utc(ct_datetime(2024, 5, 31, 23, 30)),
            "gross": Decimal("23.09"),
            "issue_date": to_utc(ct_datetime(2024, 8, 4)),
            "kwh": Decimal("302"),
            "mpan_core": "10 0000 0000 003",
            "net": Decimal("23.09"),
            "reads": [],
            "reference": "20240501_20240531_20240804_10 0000 0000 003_ssp",
            "start_date": to_utc(ct_datetime(2024, 5, 1)),
            "vat": Decimal("0.00"),
        },
        {
            "account": "10 0000 0000 003",
            "bill_type_code": "N",
            "breakdown": {
                "duos-availability-gbp": Decimal("83.12"),
                "duos-availability-kva": Decimal("41"),
                "duos-availability-rate": [Decimal("4.97")],
            },
            "finish_date": to_utc(ct_datetime(2024, 4, 30, 23, 30)),
            "gross": Decimal("83.12"),
            "issue_date": to_utc(ct_datetime(2024, 8, 4)),
            "kwh": Decimal("0"),
            "mpan_core": "10 0000 0000 003",
            "net": Decimal("83.12"),
            "reads": [],
            "reference": "20240401_20240430_20240804_10 0000 0000 "
            "003_duos-availability",
            "start_date": to_utc(ct_datetime(2024, 4, 1)),
            "vat": Decimal("0.00"),
        },
    ]
    assert bills == expected_bills


def test_make_raw_bills_october():
    wb = Workbook()
    statement_sheet = wb.worksheets[0]
    wb.create_sheet()  # HH Data
    wb.create_sheet()  # Calculations
    gduos_sheet = wb.create_sheet()  # GDUoS

    statement_sheet["D22"] = "01-Oct-2024"
    statement_sheet["F22"] = "31-Oct-2024"
    statement_sheet["H24"] = ct_datetime(2024, 8, 4)
    statement_sheet["D25"] = 1000000000003
    statement_sheet["D33"] = Decimal("302")
    statement_sheet["H33"] = Decimal("23.09")

    gduos_sheet["A2"] = 1000000000003
    gduos_sheet["B2"] = "2024-10"
    gduos_sheet["D2"] = "Capacity"
    gduos_sheet["E2"] = Decimal("41")
    gduos_sheet["F2"] = Decimal("4.97")
    gduos_sheet["G2"] = Decimal("83.12")

    bills = _make_raw_bills(wb)
    expected_bills = [
        {
            "account": "10 0000 0000 003",
            "bill_type_code": "N",
            "breakdown": {"ssp-gbp": Decimal("23.09"), "ssp-kwh": Decimal("302")},
            "finish_date": to_utc(ct_datetime(2024, 10, 31, 23, 30)),
            "gross": Decimal("23.09"),
            "issue_date": to_utc(ct_datetime(2024, 8, 4)),
            "kwh": Decimal("302"),
            "mpan_core": "10 0000 0000 003",
            "net": Decimal("23.09"),
            "reads": [],
            "reference": "20241001_20241031_20240804_10 0000 0000 003_ssp",
            "start_date": to_utc(ct_datetime(2024, 10, 1)),
            "vat": Decimal("0.00"),
        },
        {
            "account": "10 0000 0000 003",
            "bill_type_code": "N",
            "breakdown": {
                "duos-availability-gbp": Decimal("83.12"),
                "duos-availability-kva": Decimal("41"),
                "duos-availability-rate": [Decimal("4.97")],
            },
            "finish_date": to_utc(ct_datetime(2024, 10, 31, 23, 30)),
            "gross": Decimal("83.12"),
            "issue_date": to_utc(ct_datetime(2024, 8, 4)),
            "kwh": Decimal("0"),
            "mpan_core": "10 0000 0000 003",
            "net": Decimal("83.12"),
            "reads": [],
            "reference": "20241001_20241031_20240804_10 0000 0000 "
            "003_duos-availability",
            "start_date": to_utc(ct_datetime(2024, 10, 1)),
            "vat": Decimal("0.00"),
        },
    ]
    assert bills == expected_bills



import csv
from datetime import datetime as Datetime, timedelta as Timedelta
from decimal import Decimal
from io import BytesIO, StringIO
from zipfile import ZipFile

from sqlalchemy import text

from werkzeug.exceptions import BadRequest
from werkzeug.http import parse_options_header

from chellow.models import Contract, Party
from chellow.rate_server import download
from chellow.utils import hh_after, to_ct, to_utc


def _process(sess, log, set_progress, file_like):
    zip_file = ZipFile(file_like)
    name_list = zip_file.namelist()
    if len(name_list) != 1:
        raise BadRequest("The zip archive must contain exactly one file.")
    stmt = text(
        """
INSERT INTO laf (llfc_id, timestamp, value) VALUES
(unnest(CAST(:llfc_ids AS INTEGER[])), unnest(CAST(:timestamps AS TIMESTAMPTZ[])),
unnest(CAST(:values AS NUMERIC[])))
ON CONFLICT ON CONSTRAINT laf_llfc_id_timestamp_key
DO UPDATE SET (llfc_id, timestamp, value) =
(EXCLUDED.llfc_id, EXCLUDED.timestamp, EXCLUDED.value)"""
    )
    fname = name_list[0]
    csv_file = StringIO(zip_file.read(fname).decode("utf-8"))
    try:
        csv_dt = to_utc(to_ct(Datetime.strptime(fname[-12:-4], "%Y%m%d")))
    except ValueError as e:
        log(f"Can't parse the date in the file name {fname} {e}")
        return

    for llfc_ids, timestamps, values in laf_days(
        sess, log, set_progress, csv_file, csv_dt
    ):
        sess.execute(
            stmt,
            params={"llfc_ids": llfc_ids, "timestamps": timestamps, "values": values},
        )
        sess.commit()


UTC_DATETIME_MIN = to_utc(Datetime.min)


def laf_days(sess, log, set_progress, csv_file, csv_dt):
    llfc_ids = []
    timestamps = []
    values = []
    llfc_code = line_dt_ct = dno = llfc_id = None
    llfc_valid_to = UTC_DATETIME_MIN
    timestamp_cache = {}

    for line_number, vals in enumerate(csv.reader(csv_file, delimiter="|")):
        set_progress(f"Reached line number {line_number}")
        code = vals[0]

        if code == "DIS":
            participant_code = vals[1]
            dno = Party.get_by_participant_code_role_code(
                sess, participant_code, "R", csv_dt
            )

        elif code == "LLF":
            llfc_code = vals[1]
            llfc_valid_to = UTC_DATETIME_MIN

            if len(llfc_ids) > 0:
                yield llfc_ids, timestamps, values
                llfc_ids = []
                timestamps = []
                values = []

        elif code == "SDT":
            line_dt_str = vals[1]
            line_dt_ct = to_ct(Datetime.strptime(line_dt_str, "%Y%m%d"))

        elif code == "SPL":
            period, value = vals[1:]

            try:
                timestamp = timestamp_cache[line_dt_ct][period]
            except KeyError:
                try:
                    day_cache = timestamp_cache[line_dt_ct]
                except KeyError:
                    day_cache = timestamp_cache[line_dt_ct] = {}

                timestamp = day_cache[period] = to_utc(
                    line_dt_ct + Timedelta(minutes=30 * (int(period) - 1))
                )

            if hh_after(timestamp, llfc_valid_to):
                llfc = dno.find_llfc_by_code(sess, llfc_code, timestamp)
                if llfc is None:
                    continue

                llfc_id, llfc_valid_to = llfc.id, llfc.valid_to
            llfc_ids.append(llfc_id)
            timestamps.append(timestamp)
            values.append(Decimal(value))

        elif code == "ZPT":
            if len(llfc_ids) > 0:
                yield llfc_ids, timestamps, values


LAF_START = "llf"
LAF_END = "ptf.zip"


def elexon_import(sess, log, set_progress, s, scripting_key):
    log(
        "Starting to check for new LAF files for participant codes in "
        "configuration.properties.laf_importer.participant_codes"
    )
    conf = Contract.get_non_core_by_name(sess, "configuration")
    props = conf.make_properties()
    state = conf.make_state()

    try:
        laf_state = state["laf_importer"]
    except KeyError:
        laf_state = state["laf_importer"] = {}

    url = "https://downloads.elexonportal.co.uk/svallf/download"

    laf_props = props.get("laf_importer", {})
    for participant_code in laf_props.get("participant_codes", []):
        params = {"key": scripting_key, "ldso": participant_code, "format": "PTF"}

        log(
            f"Downloading from {url}?key={scripting_key}&ldso={participant_code}&"
            f"format=PTF"
        )
        sess.rollback()  # Avoid long-running transactions
        res = s.get(url, params=params)
        log(f"Received {res.status_code} {res.reason}")
        res.raise_for_status()
        try:
            cd_header = res.headers["Content-Disposition"]
        except KeyError:
            log("No Content-Disposition header found, so skipping.")
            continue

        cd = parse_options_header(cd_header)
        file_name = cd[1]["filename"]

        # File name llfetcl20220401ptf.zip
        if not file_name.startswith(LAF_START):
            raise BadRequest(f"A laf file must begin with '{LAF_START}'")
        if not file_name.endswith(LAF_END):
            raise BadRequest(f"A laf file must end with '{LAF_END}'")

        participant_code = file_name[3:7]
        timestamp = file_name[7:15]
        log(f"File name {file_name}")

        if timestamp > laf_state.get(participant_code, ""):
            data = res.content
            _process(sess, log, set_progress, BytesIO(data))
            laf_state[participant_code] = timestamp
            conf.update_state(state)
            sess.commit()
            log(f"Processed file {file_name}.")
        else:
            log(f"Already loaded {file_name}.")

    log("Finished LAF files")
    sess.commit()


def find_participant_entries(paths, laf_state):
    participant_entries = {}
    for path, url in paths:
        if len(path) == 4:
            year_str, utility, rate_type, file_name = path

            if utility == "electricity" and rate_type == "lafs":
                # File name llfetcl20220401ptf.zip
                if not file_name.startswith(LAF_START):
                    raise BadRequest(f"A laf file must begin with '{LAF_START}'")
                if not file_name.endswith(LAF_END):
                    raise BadRequest(f"A laf file must end with '{LAF_END}'")

                participant_code = file_name[3:7]
                timestamp = file_name[7:15]

                participant_state = laf_state.get(participant_code, "")
                if timestamp > participant_state:
                    try:
                        fl_entries = participant_entries[participant_code]
                    except KeyError:
                        fl_entries = participant_entries[participant_code] = {}
                    fl_entries[timestamp] = url
    return participant_entries


def rate_server_import(sess, log, set_progress, s, paths):
    log("Starting to check for new LAF files")
    conf = Contract.get_non_core_by_name(sess, "configuration")
    state = conf.make_state()

    try:
        laf_state = state["laf_importer"]
    except KeyError:
        laf_state = state["laf_importer"] = {}

    participant_entries = find_participant_entries(paths, laf_state)

    for participant_code, fl_entries in sorted(participant_entries.items()):
        log(f"Importing files for participant code {participant_code}")
        for timestamp, url in sorted(fl_entries.items()):
            log(f"Importing file with timestamp {timestamp}")

            fl = BytesIO(download(s, url))
            _process(sess, log, set_progress, fl)
            laf_state[participant_code] = timestamp
            conf.update_state(state)
            sess.commit()

    log("Finished LAF files")
    sess.commit()



from collections import defaultdict
from decimal import Decimal
from io import BytesIO


from chellow.e.bill_parsers.drax_edi import (
    Parser,
    _process_BCD,
    _process_BTL,
    _process_CCD1,
    _process_CCD3,
    _process_MAN,
    _process_MHD,
    _process_MTR,
    _process_VAT,
)
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_process_BCD(mocker):
    reference = "lkasdhhgw"
    elements = {
        "IVDT": ["200416"],
        "INVN": [reference],
        "BTCD": ["N"],
        "SUMO": ["200301", "200331"],
    }
    headers = {}
    _process_BCD(elements, headers)
    expected_headers = {
        "bill_type_code": "N",
        "finish_date": to_utc(ct_datetime(2020, 3, 31, 23, 30)),
        "issue_date": to_utc(ct_datetime(2020, 4, 16)),
        "start_date": to_utc(ct_datetime(2020, 3, 1)),
        "reference": reference,
        "elements": [],
    }
    assert headers == expected_headers


def test_process_BTL(mocker):
    elements = {"UVLT": ["7482"], "UTVA": ["7659"], "TBTL": ["8572"]}
    issue_date = to_utc(ct_datetime(2020, 4, 1))
    start_date = to_utc(ct_datetime(2020, 1, 1))
    finish_date = to_utc(ct_datetime(2020, 1, 31, 23, 30))
    headers = {
        "mpan_core": "20 0000 0000 006",
        "account": "97y4t934",
        "bill_type_code": "N",
        "reference": "974y4uot",
        "issue_date": issue_date,
        "start_date": start_date,
        "finish_date": finish_date,
        "kwh": Decimal("44.5"),
        "net": Decimal("77932.22"),
        "vat": Decimal("870937.88"),
        "gross": Decimal("986733.22"),
        "breakdown": {"raw-lines": []},
        "reads": [],
        "elements": [],
        "lines": [],
    }
    bill = _process_BTL(elements, headers)
    expected_bill = {
        "account": "97y4t934",
        "bill_type_code": "N",
        "breakdown": {"raw-lines": []},
        "finish_date": finish_date,
        "gross": Decimal("85.72"),
        "issue_date": issue_date,
        "kwh": Decimal("44.5"),
        "mpan_core": "20 0000 0000 006",
        "net": Decimal("74.82"),
        "reads": [],
        "reference": "974y4uot",
        "start_date": start_date,
        "vat": Decimal("76.59"),
        "elements": [],
    }
    assert bill == expected_bill


def test_process_BTL_decimal_places(mocker):
    elements = {"UVLT": ["7480"], "UTVA": ["7650"], "TBTL": ["8570"]}
    issue_date = to_utc(ct_datetime(2020, 4, 1))
    start_date = to_utc(ct_datetime(2020, 1, 1))
    finish_date = to_utc(ct_datetime(2020, 1, 31, 23, 30))
    headers = {
        "mpan_core": "20 0000 0000 006",
        "account": "97y4t934",
        "bill_type_code": "N",
        "reference": "974y4uot",
        "issue_date": issue_date,
        "start_date": start_date,
        "finish_date": finish_date,
        "kwh": Decimal("44.5"),
        "breakdown": {"raw-lines": []},
        "reads": [],
        "elements": [],
        "lines": [],
    }
    bill = _process_BTL(elements, headers)
    expected_net = Decimal("74.80")
    expected_vat = Decimal("76.50")
    expected_gross = Decimal("85.70")
    expected_bill = {
        "account": "97y4t934",
        "bill_type_code": "N",
        "breakdown": {"raw-lines": []},
        "finish_date": finish_date,
        "issue_date": issue_date,
        "kwh": Decimal("44.5"),
        "mpan_core": "20 0000 0000 006",
        "net": expected_net,
        "gross": expected_gross,
        "reads": [],
        "reference": "974y4uot",
        "start_date": start_date,
        "vat": expected_vat,
        "elements": [],
    }
    assert bill == expected_bill
    for actual_val, expected_val in (
        (bill["net"], expected_net),
        (bill["vat"], expected_vat),
        (bill["gross"], expected_gross),
    ):
        assert actual_val.as_tuple().exponent == expected_val.as_tuple().exponent


def test_process_CCD3():
    elements = {
        "CCDE": ["3", "", "NRG"],
        "TCOD": ["HH0002", "Night"],
        "TMOD": ["122568"],
        "CONS": ["", "", ""],
        "BPRI": ["10"],
        "CSDT": ["250401"],
        "CEDT": ["250430"],
    }

    lines = []
    headers = {
        "kwh": Decimal("0"),
        "breakdown": defaultdict(int, {}),
        "lines": lines,
        "elements": [],
    }

    _process_CCD3(elements, headers)

    expected_headers = {
        "kwh": Decimal("0"),
        "breakdown": {},
        "elements": [
            {
                "name": "nrg-msp",
                "net": Decimal("0.00"),
                "breakdown": {
                    "rate": {Decimal("0.0001")},
                },
                "start_date": to_utc(ct_datetime(2025, 4, 1)),
                "finish_date": to_utc(ct_datetime(2025, 4, 30, 23, 30)),
            }
        ],
        "lines": lines,
    }
    assert headers == expected_headers


def test_process_CCD3_multiple_elements(mocker):
    elements = {
        "CCDE": ["3", "", "NRG"],
        "TCOD": ["HH0002", "Night"],
        "TMOD": ["122568"],
        "CONS": ["8739", "KWH", ""],
        "BPRI": ["10"],
        "CSDT": ["250401"],
        "CEDT": ["250430"],
        "CTOT": ["75849", ""],
    }

    lines = []
    headers = {
        "kwh": Decimal("1000"),
        "elements": [
            {
                "name": "nrg-msp",
                "net": Decimal("1000.00"),
                "start_date": to_utc(ct_datetime(2025, 4, 1)),
                "finish_date": to_utc(ct_datetime(2025, 4, 30, 23, 30)),
                "breakdown": {
                    "rate": {Decimal("758.33")},
                    "kwh": Decimal("10000"),
                },
            },
        ],
        "breakdown": defaultdict(int, {}),
        "lines": lines,
    }

    _process_CCD3(elements, headers)

    expected_headers = {
        "breakdown": defaultdict(int, {}),
        "kwh": Decimal("1008.739"),
        "elements": [
            {
                "name": "nrg-msp",
                "net": Decimal("1000.00"),
                "start_date": to_utc(ct_datetime(2025, 4, 1)),
                "finish_date": to_utc(ct_datetime(2025, 4, 30, 23, 30)),
                "breakdown": {
                    "rate": {Decimal("758.33")},
                    "kwh": Decimal("10000"),
                },
            },
            {
                "name": "nrg-msp",
                "net": Decimal("758.49"),
                "start_date": to_utc(ct_datetime(2025, 4, 1)),
                "finish_date": to_utc(ct_datetime(2025, 4, 30, 23, 30)),
                "breakdown": {
                    "rate": {Decimal("0.0001")},
                    "kwh": Decimal("8.739"),
                },
            },
        ],
        "lines": lines,
    }
    assert headers == expected_headers


def test_process_MHD(mocker):
    message_type = "UTLBIL"
    elements = {"TYPE": [message_type]}

    customer_number = "xhgurg2"
    lines = ["xxx"]
    headers = {"customer_number": customer_number, "lines": lines}
    _process_MHD(elements, headers)
    expected_headers = {
        "customer_number": customer_number,
        "lines": lines,
        "breakdown": {"vat": {}, "raw-lines": lines},
        "kwh": Decimal(0),
        "reads": [],
    }
    assert headers == expected_headers


def test_process_MTR(mocker):
    elements = {}
    headers = {"message_type": "UTLHDR", "breakdown": {}}
    bill = _process_MTR(elements, headers)
    assert bill is None


def test_process_MAN(mocker):
    elements = {
        "MADN": ["20", "0000000000", "6", "00", "001", "002"],
    }

    headers = {}
    _process_MAN(elements, headers)
    expected_headers = {"mpan_core": "20 0000 0000 006"}
    assert headers == expected_headers


def test_process_CCD_1(mocker):
    msn = "hgkh"

    elements = {
        "CCDE": ["1", "", "NRG"],
        "TCOD": ["NIGHT", "Night"],
        "TMOD": ["453043"],
        "MTNR": [msn],
        "CONS": [[]],
        "BPRI": ["10"],
        "PRDT": ["200301"],
        "PVDT": ["200331"],
        "MLOC": ["2275834732592"],
        "PRRD": ["0", "00", "1", "00"],
        "ADJF": ["", "1"],
    }

    headers = {
        "reads": [],
        "mpan_core": "22 7583 4732 592",
    }

    _process_CCD1(elements, headers)

    expected_headers = {
        "mpan_core": "22 7583 4732 592",
        "reads": [
            {
                "msn": msn,
                "mpan": "   22 7583 4732 592",
                "coefficient": Decimal("0.00001"),
                "units": "kWh",
                "tpr_code": "00210",
                "prev_date": utc_datetime(2020, 3, 31, 22, 30),
                "prev_value": Decimal("1"),
                "prev_type_code": "N",
                "pres_date": utc_datetime(2020, 3, 1, 23, 30),
                "pres_value": Decimal("0"),
                "pres_type_code": "N",
            }
        ],
    }
    assert headers == expected_headers


def test_process_VAT(mocker):
    elements = {"UVLA": ["600"], "UVTT": ["150"], "UCSI": ["850"], "VATP": ["20000"]}
    headers = {
        "breakdown": defaultdict(int, {"vat": {}}),
    }

    _process_VAT(elements, headers)
    expected_headers = {
        "breakdown": {
            "vat-rate": {Decimal("0.20")},
            "vat": {Decimal("20"): {"net": Decimal("6.00"), "vat": Decimal("1.50")}},
        }
    }

    assert headers == expected_headers


def test_Parser(mocker, sess):
    file_lines = [
        "CDT=ZAPP+ZAPP Industries'",
        "MHD=2+UTLBIL:3'",
        "MTR=6'",
    ]
    f = BytesIO(b"\n".join(n.encode("utf8") for n in file_lines))
    parser = Parser(f)
    parser.make_raw_bills()



import json
import os
from base64 import b64decode
from datetime import datetime as Datetime
from importlib.metadata import version
from pathlib import Path

from flask import Flask, Response, g, make_response, render_template, request

from markdown_it import MarkdownIt

from markupsafe import Markup

from sqlalchemy import select

from werkzeug.exceptions import NotFound

from zish import ZishException, dumps

import chellow.api
import chellow.bank_holidays
import chellow.dloads
import chellow.e.bmarketidx
import chellow.e.bsuos
import chellow.e.elexon
import chellow.e.hh_importer
import chellow.e.rcrc
import chellow.e.system_price
import chellow.e.views
import chellow.fake_batch_updater
import chellow.gas.cv
import chellow.gas.views
import chellow.national_grid
import chellow.rrun
import chellow.testing
from chellow.models import (
    Contract,
    Session,
    User,
    UserRole,
    db_upgrade,
    start_sqlalchemy,
)
from chellow.proxy import MsProxy
from chellow.utils import HH, ct_datetime, to_ct, utc_datetime_now


TEMPLATE_FORMATS = {
    "year": "%Y",
    "month": "%m",
    "day": "%d",
    "hour": "%H",
    "minute": "%M",
    "full": "%Y-%m-%d %H:%M",
    "date": "%Y-%m-%d",
}


def get_importer_modules():
    return (
        chellow.e.elexon,
        chellow.e.bsuos,
        chellow.e.lcc,
        chellow.testing,
        chellow.bank_holidays,
        chellow.gas.cv,
        chellow.e.bmarketidx,
        chellow.national_grid,
        chellow.rate_server,
        chellow.rrun,
        chellow.fake_batch_updater,
    )


def create_app(testing=False, instance_path=None):
    app = Flask("chellow", instance_relative_config=True, instance_path=instance_path)
    app.wsgi_app = MsProxy(app.wsgi_app)
    app.secret_key = os.urandom(24)
    start_sqlalchemy()

    app.register_blueprint(chellow.views.home)
    app.register_blueprint(chellow.e.views.e)
    app.register_blueprint(chellow.gas.views.gas)

    api = chellow.api.api
    api.init_app(app, endpoint="/api/v1")

    if not testing:
        db_upgrade(app.root_path)
        chellow.e.hh_importer.startup()

        with Session() as sess:
            configuration = sess.execute(
                select(Contract).where(Contract.name == "configuration")
            ).scalar_one()
            props = configuration.make_properties()
            api_props = props.get("api", {})
            api.description = api_props.get("description", "Access Chellow data")

    chellow.dloads.startup(Path(app.instance_path), run_deleter=(not testing))

    for module in get_importer_modules():
        if not testing:
            module.startup()

    @app.before_request
    def before_request():
        g.sess = Session()

        msg = " ".join(
            "-" if v is None else v
            for v in (
                request.remote_addr,
                str(request.user_agent),
                request.remote_user,
                f"[{Datetime.now().strftime('%d/%b/%Y:%H:%M:%S')}]",
                f'"{request.method} {request.path} '
                f'{request.environ.get("SERVER_PROTOCOL")}"',
                None,
                None,
            )
        )
        print(msg)

    @app.before_request
    def check_permissions(*args, **kwargs):
        g.user = None
        config_contract = Contract.get_non_core_by_name(g.sess, "configuration")
        try:
            g.config = config_contract.make_properties()
        except ZishException:
            g.config = {}
        ad_props = g.config.get("ad_authentication", {})
        ad_auth_on = ad_props.get("on", False)
        easy_auth_props = g.config.get("easy_auth", {})
        easy_auth_on = easy_auth_props.get("on", False)
        path = request.path
        if path in ("/health", "/robots933456.txt"):
            return
        if ad_auth_on:
            username = request.headers["X-Isrw-Proxy-Logon-User"].upper()
            user = g.sess.scalars(
                select(User).where(User.email_address == username)
            ).first()
            if user is None:
                try:
                    username = ad_props["default_user"]
                    user = g.sess.scalars(
                        select(User).where(User.email_address == username)
                    ).first()
                except KeyError:
                    user = None
            if user is not None:
                g.user = user
        elif easy_auth_on:
            jwt_enc = request.headers["X-MS-CLIENT-PRINCIPAL"]
            jwt = json.loads(b64decode(jwt_enc))
            for claim in jwt["claims"]:
                if claim["typ"] == (
                    "http://schemas.xmlsoap.org/ws/2005/05/identity/claims/"
                    "emailaddress"
                ):
                    username = claim["val"]
                    user = g.sess.scalars(
                        select(User).where(User.email_address == username)
                    ).first()
                    break
            if user is None:
                try:
                    username = easy_auth_props["default_user"]
                    user = g.sess.scalars(
                        select(User).where(User.email_address == username)
                    ).first()
                except KeyError:
                    user = None
            if user is not None:
                g.user = user
        else:
            auth = request.authorization

            if auth is None:
                try:
                    ips = g.config["ips"]
                    if request.remote_addr in ips:
                        key = request.remote_addr
                    elif "*.*.*.*" in ips:
                        key = "*.*.*.*"
                    else:
                        key = None

                    email = ips[key]
                    g.user = g.sess.scalars(
                        select(User).where(User.email_address == email)
                    ).first()
                except KeyError:
                    pass
            else:
                user = g.sess.scalars(
                    select(User).where(User.email_address == auth.username)
                ).first()
                if user is not None and user.password_matches(auth.password):
                    g.user = user

        # Got our user
        method = request.method

        if g.user is not None:
            if "X-Isrw-Proxy-Logon-User" in request.headers:
                g.user.proxy_username = request.headers[
                    "X-Isrw-Proxy-Logon-User"
                ].upper()

            role = g.user.user_role
            role_code = role.code

            if (
                role_code == "viewer"
                and (
                    method in ("GET", "HEAD")
                    or path
                    in (
                        "/reports/59",
                        "/reports/169",
                        "/reports/183",
                        "/reports/187",
                        "/reports/247",
                        "/reports/111",
                        "/reports/149",
                    )
                )
                and path not in ("/system",)
            ):
                return
            elif role_code == "editor":
                return
            elif role_code == "party-viewer":
                if method in ("GET", "HEAD"):
                    party = g.user.party
                    market_role_code = party.market_role.code
                    if market_role_code in ("C", "D"):
                        dc_contract_id = request.args["dc_contract_id"]
                        dc_contract = Contract.get_dc_by_id(g.sess, dc_contract_id)
                        if dc_contract.party == party and request.full_path.startswith(
                            "/channel_snags?"
                        ):
                            return
                    elif market_role_code == "X":
                        if path.startswith(f"/e/supplier_contracts/{party.id}"):
                            return

        if g.user is None and g.sess.query(User).count() == 0:
            g.sess.rollback()
            user_role = g.sess.query(UserRole).filter(UserRole.code == "editor").one()
            User.insert(g.sess, "admin@example.com", "admin", user_role, None)
            g.sess.commit()
            return

        if (not easy_auth_on) and (
            (g.user is None) or (not ad_auth_on and auth is None)
        ):
            return Response(
                "Could not verify your access level for that URL.\n"
                "You have to login with proper credentials",
                401,
                {"WWW-Authenticate": 'Basic realm="Chellow"'},
            )

        return make_response(render_template("403.html", properties=g.config), 403)

    @app.context_processor
    def chellow_context_processor():
        global_alerts = []
        for task in chellow.e.hh_importer.tasks.values():
            if task.is_error:
                try:
                    contract = Contract.get_by_id(g.sess, task.contract_id)
                    global_alerts.append(
                        f"There's a problem with the <a "
                        f"href='/e/dc_contracts/{contract.id}/auto_importer'>"
                        f"automatic HH data importer for contract '{contract.name}'"
                        f"</a>.",
                    )
                except NotFound:
                    pass

        for mod in get_importer_modules():
            importer = mod.get_importer()
            if importer is not None and importer.global_alert is not None:
                global_alerts.append(importer.global_alert)

        test_match = g.config.get("test_match")

        def get_month_max_day(year, month):
            if month == 12:
                new_year = year + 1
                new_month = 1
            else:
                new_year = year
                new_month = month + 1
            return (ct_datetime(new_year, new_month) - HH).day

        return {
            "current_user": g.user,
            "global_alerts": global_alerts,
            "is_test": False if test_match is None else test_match in request.host,
            "test_css": g.config.get("test_css"),
            "get_month_max_day": get_month_max_day,
        }

    @app.teardown_request
    def shutdown_session(exception=None):
        if getattr(g, "sess", None) is not None:
            g.sess.close()

    @app.template_filter("hh_format")
    def hh_format_filter(dt, modifier="full"):
        if dt is None:
            return "Ongoing"
        else:
            return to_ct(dt).strftime(TEMPLATE_FORMATS[modifier])

    @app.template_filter("now_if_none")
    def now_if_none(dt):
        return utc_datetime_now() if dt is None else dt

    @app.template_filter("to_ct")
    def to_ct_filter(dt):
        return to_ct(dt)

    @app.template_filter("dumps")
    def dumps_filter(d):
        return dumps(d)

    @app.template_filter("markdown")
    def markdown_filter(txt):
        md = MarkdownIt()
        return Markup(md.render(txt))

    return app


__version__ = version("chellow")



import csv
from io import StringIO

import chellow.dloads
from chellow.models import (
    Contract,
    MarketRole,
    Participant,
    ReportRun,
    Site,
    User,
    UserRole,
)
from chellow.reports.report_asset_comparison import _process_sites, content
from chellow.utils import ct_datetime, to_utc


def test_content(sess):
    editor = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "xxx", editor, None)
    user_id = user.id
    report_run = ReportRun.insert(sess, "asset", None, "asset", {})
    vf = to_utc(ct_datetime(2020, 1, 1))
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    Contract.insert_non_core(
        sess,
        "configuration",
        "",
        {"asset_comparison": {"ignore_site_codes": []}},
        vf,
        None,
        {},
    )
    sess.commit()

    asset_str = "\n".join(("code,,,status",))
    file_like = StringIO(asset_str)
    report_run_id = report_run.id
    content(user_id, file_like, report_run_id)

    files = list(p.name for p in chellow.dloads.download_path.iterdir())
    assert files == ["00000_FINISHED_admin_example_com_asset_comparison.csv"]


def test_process_sites(sess):
    site_code = "109"
    Site.insert(sess, site_code, "A Site")
    report_run = ReportRun.insert(sess, "asset", None, "asset", {})
    sess.commit()

    asset_str = "\n".join(("code,,,status",))
    file_like = StringIO(asset_str)
    output = StringIO()
    writer = csv.writer(output, lineterminator="\n")
    props = {"asset_comparison": {"ignore_site_codes": []}}

    _process_sites(sess, file_like, writer, props, report_run.id)

    assert output.getvalue() == "site_code,asset_status,chellow_status\n"



import collections
import importlib
import threading
import traceback
from io import BytesIO
from pkgutil import iter_modules

from sqlalchemy import select

from werkzeug.exceptions import BadRequest

import chellow
import chellow.e.bill_parsers.engie_xls
from chellow.models import (
    Batch,
    BatchFile,
    BillType,
    Contract,
    ReadType,
    Session,
    Supply,
    Tpr,
)
from chellow.utils import ct_datetime_now, keydefaultdict


import_id = 0
import_lock = threading.Lock()
imports = {}


def find_parser_names():
    return [name for _, name, _ in iter_modules(chellow.e.bill_parsers.__path__)]


class BillImport(threading.Thread):
    def __init__(self, batch_contract, now=None):
        threading.Thread.__init__(self)
        global import_id
        self.import_id = import_id
        import_id += 1

        if isinstance(batch_contract, Batch):
            self.contract_id = None
            self.batch_id = batch_contract.id
        elif isinstance(batch_contract, Contract):
            self.contract_id = batch_contract.id
            self.batch_id = None
        else:
            raise BadRequest("batch_contract must be a Batch or Contract.")
        self.successful_bills = []
        self.failed_bills = []
        self.log = collections.deque()
        self.bill_num = None
        self.parser = None
        self.now = now

    def _log(self, msg):
        with import_lock:
            ts = ct_datetime_now() if self.now is None else self.now
            self.log.appendleft(f"{ts.strftime('%Y-%m-%d %H:%M:%S')} - {msg}")

    def status(self):
        if self.is_alive():
            if self.bill_num is not None:
                return f"Inserting raw bills: I've reached bill number {self.bill_num}."
            elif self.parser is not None:
                return (
                    f"Parsing file: I've reached line number {self.parser.line_number}."
                )
            else:
                return "Running"
        else:
            return "Not running"

    def run(self):
        try:
            batch_ids = []
            with Session() as sess:
                batch_q = select(Batch).order_by(Batch.reference)
                if self.batch_id is not None:
                    batch = Batch.get_by_id(sess, self.batch_id)
                    batch_q = batch_q.where(Batch.id == self.batch_id)
                elif self.contract_id is not None:
                    contract = Contract.get_by_id(sess, self.contract_id)
                    batch_q = batch_q.where(Batch.contract == contract)
                for batch in sess.scalars(batch_q):
                    batch_ids.append(batch.id)

            for batch_id in batch_ids:
                with Session() as sess:
                    batch = Batch.get_by_id(sess, batch_id)
                    self._log(f"Importing bills from batch {batch.reference}.")
                    bf_q = (
                        select(BatchFile)
                        .where(BatchFile.batch == batch)
                        .order_by(BatchFile.upload_timestamp)
                    )
                    bill_types = keydefaultdict(lambda k: BillType.get_by_code(sess, k))

                    tprs = keydefaultdict(
                        lambda k: None if k is None else Tpr.get_by_code(sess, k)
                    )

                    read_types = keydefaultdict(lambda k: ReadType.get_by_code(sess, k))

                    for bf in sess.scalars(bf_q):
                        self.parser = _process_batch_file(sess, bf, self._log)
                        for self.bill_num, raw_bill in enumerate(
                            self.parser.make_raw_bills()
                        ):
                            batch = bf.batch
                            sum_elem = sum(el["net"] for el in raw_bill["elements"])
                            raw_bill_net = raw_bill["net"]
                            if sum_elem != raw_bill_net:
                                raw_bill["error"] = (
                                    f"The sum of the elements' net {sum_elem} doesn't "
                                    f"equal the bill net {raw_bill_net}."
                                )
                            if "error" in raw_bill:
                                self.failed_bills.append(raw_bill)
                            else:
                                try:
                                    mpan_core = raw_bill["mpan_core"]
                                    supply = Supply.get_by_mpan_core(sess, mpan_core)
                                    with sess.begin_nested():
                                        bill = batch.insert_bill(
                                            sess,
                                            raw_bill["account"],
                                            raw_bill["reference"],
                                            raw_bill["issue_date"],
                                            raw_bill["start_date"],
                                            raw_bill["finish_date"],
                                            raw_bill["kwh"],
                                            raw_bill_net,
                                            raw_bill["vat"],
                                            raw_bill["gross"],
                                            bill_types[raw_bill["bill_type_code"]],
                                            raw_bill["breakdown"],
                                            supply,
                                        )
                                        for raw_element in raw_bill["elements"]:
                                            bill.insert_element(
                                                sess,
                                                raw_element["name"],
                                                raw_element["start_date"],
                                                raw_element["finish_date"],
                                                raw_element["net"],
                                                raw_element["breakdown"],
                                            )
                                        for raw_read in raw_bill["reads"]:
                                            bill.insert_read(
                                                sess,
                                                tprs[raw_read["tpr_code"]],
                                                raw_read["coefficient"],
                                                raw_read["units"],
                                                raw_read["msn"],
                                                raw_read["mpan"],
                                                raw_read["prev_date"],
                                                raw_read["prev_value"],
                                                read_types[raw_read["prev_type_code"]],
                                                raw_read["pres_date"],
                                                raw_read["pres_value"],
                                                read_types[raw_read["pres_type_code"]],
                                            )
                                        self.successful_bills.append(raw_bill)
                                except KeyError as e:
                                    err = raw_bill.get("error", "")
                                    raw_bill["error"] = err + " " + str(e)
                                    self.failed_bills.append(raw_bill)
                                except BadRequest as e:
                                    raw_bill["error"] = str(e.description)
                                    self.failed_bills.append(raw_bill)

                    if len(self.failed_bills) == 0:
                        sess.commit()
                        self._log(
                            "All the bills have been successfully loaded and attached "
                            "to the batch."
                        )
                    else:
                        sess.rollback()
                        self._log(
                            f"The import has finished, but there were "
                            f"{len(self.failed_bills)} "
                            f"failures, and so the whole import has been rolled back."
                        )
                        break

        except BadRequest as e:
            msg = f"Problem: {e.description}"
            if e.__cause__ is not None:
                msg += f" {traceback.format_exc()}"
            self._log(msg)
        except BaseException:
            self._log(f"I've encountered a problem: {traceback.format_exc()}")

    def make_fields(self):
        with import_lock:
            fields = {
                "log": tuple(self.log),
                "is_alive": self.is_alive(),
                "importer_id": self.import_id,
            }
            if not self.is_alive():
                fields["successful_bills"] = self.successful_bills
                fields["failed_bills"] = self.failed_bills
            return fields


def _process_batch_file(sess, batch_file, log_f):
    data = batch_file.data
    parser_name = batch_file.parser_name

    if len(data) == 0:
        raise BadRequest("File has zero length")

    try:
        imp_mod = importlib.import_module(f"chellow.e.bill_parsers.{parser_name}")
    except ImportError:
        raise BadRequest(f"Can't find a parser with the name '{parser_name}'.")

    parser = imp_mod.Parser(BytesIO(data))
    log_f(f"Starting to parse the file {batch_file.filename} with '{parser_name}'.")

    return parser


def start_bill_import(batch):
    with import_lock:
        bi = BillImport(batch)
        imports[bi.import_id] = bi
        bi.start()

    return bi.import_id


def start_bill_import_contract(contract):
    with import_lock:
        bi = BillImport(contract)
        imports[bi.import_id] = bi
        bi.start()

    return bi.import_id


def get_bill_import_ids(batch):
    with import_lock:
        return [k for k, v in imports.items() if v.batch_id == batch.id]


def get_bill_import_ids_contract(contract):
    with import_lock:
        return [k for k, v in imports.items() if v.contract_id == contract.id]


def get_bill_import(id):
    with import_lock:
        return imports[id]



import csv

from io import StringIO

from sqlalchemy import select

from utils import match, match_tables

from chellow.models import (
    Comm,
    Contract,
    Cop,
    DtcMeterType,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcParticipant,
    Participant,
    Pc,
    Site,
    Source,
    User,
    UserRole,
    VoltageLevel,
    insert_comms,
    insert_cops,
    insert_dtc_meter_types,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.reports.report_channel_snags import content
from chellow.utils import ct_datetime, to_utc


def test_do_get_as_csv(mocker, sess, client):
    mock_Thread = mocker.patch("chellow.reports.report_channel_snags.threading.Thread")
    vf = to_utc(ct_datetime(2000, 1, 1))
    market_role_Z = MarketRole.get_by_code(sess, "Z")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    user = User.get_by_email_address(sess, "admin@example.com")
    user_id = user.id
    dc_contract_id = dc_contract.id
    sess.commit()
    f = StringIO()
    mocker.patch("chellow.reports.report_channel_snags.open_file", return_value=f)
    days_hidden = 0
    is_ignored = True
    show_settlement = "both"
    only_ongoing = True
    as_csv = True
    days_long_hidden = ""
    now = to_utc(ct_datetime(2025, 10, 31, 0, 0))
    query_string = {
        "days_hidden": days_hidden,
        "show_settlement": show_settlement,
        "is_ignored": is_ignored,
        "only_ongoing": only_ongoing,
        "dc_contract_id": dc_contract_id,
        "as_csv": as_csv,
        "days_long_hidden": days_long_hidden,
        "now_year": 2025,
        "now_month": 10,
        "now_day": 31,
        "now_hour": 0,
        "now_minute": 0,
        "comm_id": 0,
    }
    response = client.get("/reports/channel_snags", query_string=query_string)
    match(response, 303)

    args = (
        dc_contract_id,
        days_hidden,
        is_ignored,
        user_id,
        only_ongoing,
        show_settlement,
        None,
        now,
        [0],
    )
    mock_Thread.assert_called_with(target=content, args=args)


def test_do_get_as_html(mocker, sess, client):
    vf = to_utc(ct_datetime(2000, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.get_by_code(sess, "Z")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    dc_contract_id = dc_contract.id
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess, "bank_holidays", "", {}, vf, None, bank_holiday_rate_script
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)

    mop_contract = Contract.insert_mop(
        sess, "Fusion Mop Contract", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    Contract.insert_dno(sess, dno.dno_code, participant, "", {}, vf, None, {})
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        to_utc(ct_datetime(2000, 1, 1)),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    era = supply.eras[0]
    era.insert_channel(sess, True, "ACTIVE")
    sess.commit()

    days_hidden = 0
    is_ignored = True
    show_settlement = "both"
    only_ongoing = True
    days_long_hidden = ""
    query_string = {
        "days_hidden": days_hidden,
        "show_settlement": show_settlement,
        "is_ignored": is_ignored,
        "only_ongoing": only_ongoing,
        "dc_contract_id": dc_contract_id,
        "days_long_hidden": days_long_hidden,
    }
    response = client.get("/reports/channel_snags", query_string=query_string)
    match(response, 200)


def test_content(mocker, sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    editor = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "xxx", editor, None)
    user_id = user.id
    dc_contract_id = dc_contract.id

    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess, "bank_holidays", "", {}, vf, None, bank_holiday_rate_script
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)

    mop_contract = Contract.insert_mop(
        sess, "Fusion Mop Contract", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    Contract.insert_dno(sess, dno.dno_code, participant, "", {}, vf, None, {})
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        to_utc(ct_datetime(2000, 1, 1)),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    era = supply.eras[0]
    era.insert_channel(sess, True, "ACTIVE")
    comm_ids = [c.id for c in sess.scalars(select(Comm))]
    sess.commit()
    f = StringIO()
    f.close = mocker.Mock()
    mocker.patch("chellow.reports.report_channel_snags.open_file", return_value=f)
    days_hidden = 0
    is_ignored = True
    days_since_finished = ""
    show_settlement = "both"
    days_long_hidden = None
    now = to_utc(ct_datetime(2025, 10, 30, 0, 0))
    content(
        dc_contract_id,
        days_hidden,
        is_ignored,
        user_id,
        days_since_finished,
        show_settlement,
        days_long_hidden,
        now,
        comm_ids,
    )
    f.seek(0)
    actual = list(csv.reader(f))
    expected = [
        [
            "contract",
            "hidden_days",
            "chellow_ids",
            "imp_mpan_core",
            "exp_mpan_core",
            "site_code",
            "site_name",
            "snag_description",
            "channel_types",
            "comm",
            "start_date",
            "finish_date",
            "is_ignored",
            "days_since_finished",
            "duration",
        ],
        [
            "Fusion DC 2000",
            "0",
            "1",
            "22 7867 6232 781",
            "",
            "CI017",
            "Water Works",
            "Missing",
            "True_ACTIVE",
            "GSM",
            "2000-01-01 00:00",
            "",
            "False",
            "",
            "9434",
        ],
    ]
    match_tables(expected, actual)


def test_content_not_show_settlement(mocker, sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    editor = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "xxx", editor, None)
    user_id = user.id
    dc_contract_id = dc_contract.id
    insert_comms(sess)
    comm_ids = [c.id for c in sess.scalars(select(Comm))]
    sess.commit()

    f = StringIO()
    f.close = mocker.Mock()
    mocker.patch("chellow.reports.report_channel_snags.open_file", return_value=f)
    days_hidden = 0
    is_ignored = True
    days_since_finished = ""
    show_settlement = "both"
    days_long_hidden = ""
    now = to_utc(ct_datetime(2025, 10, 30, 0, 0))
    content(
        dc_contract_id,
        days_hidden,
        is_ignored,
        user_id,
        days_since_finished,
        show_settlement,
        days_long_hidden,
        now,
        comm_ids,
    )
    f.seek(0)
    actual = list(csv.reader(f))
    expected = [
        [
            "contract",
            "hidden_days",
            "chellow_ids",
            "imp_mpan_core",
            "exp_mpan_core",
            "site_code",
            "site_name",
            "snag_description",
            "channel_types",
            "comm",
            "start_date",
            "finish_date",
            "is_ignored",
            "days_since_finished",
            "duration",
        ]
    ]
    match_tables(expected, actual)


def test_content_days_long_hidden(mocker, sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    editor = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "xxx", editor, None)
    user_id = user.id
    dc_contract_id = dc_contract.id

    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess, "bank_holidays", "", {}, vf, None, bank_holiday_rate_script
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)

    mop_contract = Contract.insert_mop(
        sess, "Fusion Mop Contract", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    Contract.insert_dno(sess, dno.dno_code, participant, "", {}, vf, None, {})
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        to_utc(ct_datetime(2000, 1, 1)),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    era = supply.eras[0]
    era.insert_channel(sess, True, "ACTIVE")
    comm_ids = [c.id for c in sess.scalars(select(Comm))]
    sess.commit()
    f = StringIO()
    f.close = mocker.Mock()
    mocker.patch("chellow.reports.report_channel_snags.open_file", return_value=f)
    days_hidden = 0
    is_ignored = True
    days_since_finished = ""
    show_settlement = "both"
    days_long_hidden = 1
    now = to_utc(ct_datetime(2025, 10, 30, 0, 0))
    content(
        dc_contract_id,
        days_hidden,
        is_ignored,
        user_id,
        days_since_finished,
        show_settlement,
        days_long_hidden,
        now,
        comm_ids,
    )
    f.seek(0)
    actual = list(csv.reader(f))
    expected = [
        [
            "contract",
            "hidden_days",
            "chellow_ids",
            "imp_mpan_core",
            "exp_mpan_core",
            "site_code",
            "site_name",
            "snag_description",
            "channel_types",
            "comm",
            "start_date",
            "finish_date",
            "is_ignored",
            "days_since_finished",
            "duration",
        ],
        [
            "Fusion DC 2000",
            "0",
            "1",
            "22 7867 6232 781",
            "",
            "CI017",
            "Water Works",
            "Missing",
            "True_ACTIVE",
            "GSM",
            "2000-01-01 00:00",
            "",
            "False",
            "",
            "9434",
        ],
    ]
    match_tables(expected, actual)



import csv
from decimal import Decimal, InvalidOperation
from io import StringIO

from dateutil.relativedelta import relativedelta

from werkzeug.exceptions import BadRequest

from zish import ZishLocationException

from chellow.utils import (
    HH,
    ct_datetime_parse,
    loads,
    parse_mpan_core,
    to_utc,
    validate_hh_start,
)


def parse_date(row, idx, is_finish=False):
    date_str = row[idx].strip()
    try:
        if len(date_str) == 10:
            dt = to_utc(ct_datetime_parse(date_str, "%Y-%m-%d"))
            if is_finish:
                dt = dt + relativedelta(days=1) - HH
        else:
            dt = to_utc(ct_datetime_parse(date_str, "%Y-%m-%d %H:%M"))
        return validate_hh_start(dt)
    except BaseException as e:
        raise BadRequest(
            f"Can't parse the date at column {idx}. The required spreadsheet format is "
            f"'YYYY-MM-DD HH:MM'. {e}"
        )


def to_decimal(vals, dec_index, dec_name, is_money=False):
    try:
        dec_str = vals[dec_index]
        dec_str = dec_str.replace(",", "")
        dec = Decimal(dec_str)
        if is_money:
            dec += Decimal("0.00")
            dec = round(dec, 2)
        return dec
    except InvalidOperation as e:
        raise BadRequest(
            f"The value '{dec_str}' can't be parsed as a decimal: {e}. It's in "
            f"the '{dec_name}' column at position {dec_index} in {vals}."
        )
    except IndexError:
        raise BadRequest(
            f"The field '{dec_name}' can't be found. It's expected at "
            f"position {dec_index} in the list of fields."
        )
    except ValueError:
        raise BadRequest(
            f"The {dec_name} field '{dec_str}' cannot be parsed as a number. The "
            f"{dec_name} field is the {dec_index} field of {vals}."
        )


def _process_row(vals):
    bill_type_code = vals[0]
    account = vals[1]
    mpan_core = parse_mpan_core(vals[2])
    reference = vals[3]
    issue_date = parse_date(vals, 4)
    start_date = parse_date(vals, 5)
    finish_date = parse_date(vals, 6, True)

    kwh = to_decimal(vals, 7, "kwh")
    net = to_decimal(vals, 8, "net", is_money=True)
    vat = to_decimal(vals, 9, "vat", is_money=True)
    gross = to_decimal(vals, 10, "gross", is_money=True)

    if len(vals) > 11:
        breakdown_str = vals[11].strip()
        if len(breakdown_str) == 0:
            breakdown = {}
        else:
            try:
                breakdown = loads(breakdown_str)
            except ZishLocationException as e:
                raise BadRequest(str(e))
    else:
        raise BadRequest("There isn't a 'breakdown' field on the end.")

    while vals[-1] == "" and len(vals) > 12:
        del vals[-1]

    reads = []
    elements = []
    i = 12
    while i < len(vals):
        typ = vals[i].strip().lower()
        if typ == "read":
            tpr_str = vals[i + 5].strip()
            tpr_code = None if len(tpr_str) == 0 else tpr_str.zfill(5)
            reads.append(
                {
                    "msn": vals[i + 1],
                    "mpan": vals[i + 2],
                    "coefficient": to_decimal(vals, i + 3, "coefficient"),
                    "units": vals[i + 4],
                    "tpr_code": tpr_code,
                    "prev_date": parse_date(vals, i + 6),
                    "prev_value": Decimal(vals[i + 7]),
                    "prev_type_code": vals[i + 8],
                    "pres_date": parse_date(vals, i + 9),
                    "pres_value": Decimal(vals[i + 10]),
                    "pres_type_code": vals[i + 11],
                }
            )
            i += 12
        elif typ == "element":
            breakdown_str = vals[i + 5].strip()
            if len(breakdown_str) == 0:
                breakdown = {}
            else:
                try:
                    breakdown = loads(breakdown_str)
                except ZishLocationException as e:
                    raise BadRequest(str(e))
            elements.append(
                {
                    "name": vals[i + 1],
                    "start_date": parse_date(vals, i + 2),
                    "finish_date": parse_date(vals, i + 3),
                    "net": to_decimal(vals, i + 4, "net", is_money=True),
                    "breakdown": breakdown,
                }
            )
            i += 6
        else:
            raise BadRequest("Record type {type} not recognized.")

    if len(elements) == 0:
        elements.append(
            {
                "name": "net",
                "start_date": start_date,
                "finish_date": finish_date,
                "net": net,
                "breakdown": {},
            }
        )

    return {
        "bill_type_code": bill_type_code,
        "account": account,
        "mpan_core": mpan_core,
        "reference": reference,
        "issue_date": issue_date,
        "start_date": start_date,
        "finish_date": finish_date,
        "kwh": kwh,
        "net": net,
        "vat": vat,
        "gross": gross,
        "breakdown": breakdown,
        "reads": reads,
        "elements": elements,
    }


class Parser:
    def __init__(self, f):
        self.reader = csv.reader(
            StringIO(str(f.read(), "utf-8-sig", errors="ignore")), skipinitialspace=True
        )
        self.line_number = None

    def make_raw_bills(self):
        for self.line_number, self.vals in enumerate(self.reader, start=2):
            try:
                # skip blank lines and comment lines
                if (
                    len(self.vals) == 0
                    or set(self.vals) == {""}
                    or self.vals[0].strip().startswith("#")
                ):
                    continue
                yield _process_row(self.vals)
            except BadRequest as e:
                raise BadRequest(
                    f"Problem at line {self.line_number} {self.vals}: {e.description}"
                )



import csv
from decimal import Decimal

from dateutil.relativedelta import relativedelta


from werkzeug.exceptions import BadRequest

from zish import loads

from chellow.models import Contract, RateScript
from chellow.utils import (
    ct_datetime_parse,
    hh_format,
    to_utc,
    u_months_u,
    utc_datetime_now,
)


ELEXON_PORTAL_SCRIPTING_KEY_KEY = "elexonportal_scripting_key"


def hh(data_source):
    try:
        cache = data_source.caches["rcrc"]
    except KeyError:
        cache = data_source.caches["rcrc"] = {}

    for hh in data_source.hh_data:
        try:
            hh["rcrc-rate"] = rcrc = cache[hh["start-date"]]
        except KeyError:
            h_start = hh["start-date"]
            rates = data_source.non_core_rate("rcrc", h_start)["rates"]
            try:
                hh["rcrc-rate"] = rcrc = cache[h_start] = (
                    float(rates[key_format(h_start)]) / 1000
                )
            except KeyError:
                try:
                    dt = h_start - relativedelta(days=3)
                    hh["rcrc-rate"] = rcrc = cache[h_start] = (
                        float(rates[key_format(dt)]) / 1000
                    )
                except KeyError:
                    raise BadRequest(
                        f"For the RCRC rate script at {hh_format(dt)} the rate cannot "
                        f"be found."
                    )

        hh["rcrc-kwh"] = hh["nbp-kwh"]
        hh["rcrc-gbp"] = hh["nbp-kwh"] * rcrc


def key_format(dt):
    return dt.strftime("%d %H:%M Z")


def _find_month(lines, month_start, month_finish):
    parser = csv.reader(lines, delimiter=",", quotechar='"')
    next(parser)
    next(parser)
    month_rcrcs = {}
    for values in parser:
        if len(values) == 0:
            continue
        date_str = values[0]
        if "/" in date_str:
            pattern = "%d/%m/%Y"
        elif " " in date_str:
            pattern = "%d %b %Y"
        else:
            raise BadRequest(f"Date format {date_str} not recognized.")
        hh_date = to_utc(ct_datetime_parse(date_str, pattern))
        hh_date += relativedelta(minutes=30 * int(values[2]))
        if month_start <= hh_date <= month_finish:
            month_rcrcs[key_format(hh_date)] = Decimal(values[3])
    return month_rcrcs


def elexon_import(sess, log, set_progress, s, scripting_key):
    log("Starting to check RCRCs.")
    contract = Contract.get_non_core_by_name(sess, "rcrc")
    latest_rs = (
        sess.query(RateScript)
        .filter(RateScript.contract_id == contract.id)
        .order_by(RateScript.start_date.desc())
        .first()
    )
    latest_rs_id = latest_rs.id
    latest_rs_start = latest_rs.start_date

    months = list(
        u_months_u(
            start_year=latest_rs_start.year, start_month=latest_rs_start.month, months=2
        )
    )
    month_start, month_finish = months[1]
    now = utc_datetime_now()
    if now > month_finish:
        params = {"key": scripting_key}
        url_str = "https://downloads.elexonportal.co.uk/file/download/RCRC_FILE"
        log(
            f"Downloading {url_str}?key={scripting_key} to see if data is available "
            f"from {hh_format(month_start)} to {hh_format(month_finish)}."
        )

        sess.rollback()  # Avoid long-running transaction
        r = s.get(url_str, timeout=120, params=params)
        month_rcrcs = _find_month(
            (x.decode() for x in r.iter_lines()), month_start, month_finish
        )
        if key_format(month_finish) in month_rcrcs:
            log("The whole month's data is there.")
            script = {"rates": month_rcrcs}
            contract = Contract.get_non_core_by_name(sess, "rcrc")
            rs = RateScript.get_by_id(sess, latest_rs_id)
            contract.update_rate_script(
                sess, rs, rs.start_date, month_finish, loads(rs.script)
            )
            contract.insert_rate_script(sess, month_start, script)
            sess.commit()
            log(f"Added a new rate script starting at {hh_format(month_start)}.")
        else:
            msg = "There isn't a whole month there yet."
            if len(month_rcrcs) > 0:
                msg += f" The last date is {sorted(month_rcrcs.keys())[-1]}"
            log(msg)



def vb(ds):

    for hh in ds.hh_data:
        start_date = hh["start_date"]
        soq = hh["soq"]
        aq = hh["aq"]
        nts_rates = ds.g_industry_rates("nts_commodity", start_date)

        to_exit_commodity_rate = float(nts_rates["to_exit_gbp_per_kwh"])
        hh["to_exit_commodity_rate"] = to_exit_commodity_rate

        so_exit_commodity_rate = float(nts_rates["so_exit_gbp_per_kwh"])
        hh["so_exit_commodity_rate"] = so_exit_commodity_rate

        rates = ds.g_industry_rates("dn", start_date)
        dn_rates = rates["gdn"][ds.g_dn_code]
        system_commodity_rates = dn_rates["system_commodity"]
        system_capacity_rates = dn_rates["system_capacity"]
        customer_capacity_rates = dn_rates["customer_capacity"]
        customer_fixed_rates = dn_rates["customer_fixed"]

        if aq <= 73200:
            pref = "to_73200_"
            dn_system_commodity_rate = float(
                system_commodity_rates[pref + "gbp_per_kwh"]
            )
            dn_system_capacity_rate = float(
                system_capacity_rates[pref + "gbp_per_kwh_per_day"]
            )
            dn_customer_capacity_rate = float(
                customer_capacity_rates[pref + "gbp_per_kwh_per_day"]
            )
            dn_customer_fixed_rate = 0
        elif 73200 < aq < 732000:
            pref = "73200_to_732000_"
            dn_system_commodity_rate = float(
                system_commodity_rates[pref + "gbp_per_kwh"]
            )
            dn_system_capacity_rate = float(
                system_capacity_rates[pref + "gbp_per_kwh_per_day"]
            )
            dn_customer_capacity_rate = float(
                customer_capacity_rates[pref + "gbp_per_kwh_per_day"]
            )
            if ds.g_reading_frequency_code == "A":
                dn_customer_fixed_rate = float(
                    customer_fixed_rates[pref + "biannual_gbp_per_day"]
                )
            else:
                dn_customer_fixed_rate = float(
                    customer_fixed_rates[pref + "monthly_gbp_per_day"]
                )

        else:
            key = "732000_and_over"

            system_commodity_rts = system_commodity_rates[key]
            dn_system_commodity_rate = max(
                float(system_commodity_rts["minimum_gbp_per_kwh"]),
                float(system_commodity_rts["gbp_per_kwh"])
                * soq ** float(system_commodity_rts["exponent"]),
            )

            system_capacity_rts = system_capacity_rates[key]
            dn_system_capacity_rate = max(
                float(system_capacity_rts["minimum_gbp_per_kwh_per_day"]),
                float(system_capacity_rts["gbp_per_kwh_per_day"])
                * soq ** float(system_capacity_rts["exponent"]),
            )

            customer_capacity_rts = customer_capacity_rates[key]
            dn_customer_capacity_rate = float(
                customer_capacity_rts["gbp_per_kwh_per_day"]
            ) * soq ** float(customer_capacity_rts["exponent"])

            dn_customer_fixed_rate = 0

        hh["dn_system_commodity_rate"] = dn_system_commodity_rate

        hh["dn_system_capacity_rate"] = dn_system_capacity_rate

        dn_system_capacity_fixed_rate = soq * dn_system_capacity_rate
        hh["dn_system_capacity_fixed_rate"] = dn_system_capacity_fixed_rate

        hh["dn_customer_capacity_rate"] = dn_customer_capacity_rate

        dn_customer_capacity_fixed_rate = soq * dn_customer_capacity_rate
        hh["dn_customer_capacity_fixed_rate"] = dn_customer_capacity_fixed_rate

        hh["dn_customer_fixed_rate"] = dn_customer_fixed_rate

        dn_ecn_rate = float(
            rates["exit_zones"][ds.g_exit_zone_code][
                "exit_capacity_gbp_per_kwh_per_day"
            ]
        )
        hh["dn_ecn_rate"] = dn_ecn_rate

        dn_ecn_fixed_rate = dn_ecn_rate * soq
        hh["dn_ecn_fixed_rate"] = dn_ecn_fixed_rate
